#!/bin/bash

# Define default thresholds and namespaces if not provided
DEFAULT_AGE_THRESHOLD=180 # Age threshold in minutes for SparkApplication
POD_AGE_THRESHOLD=2880 # 2 days for a pod
POD_TERMINATING_STATE_THRESHOLD=5 # if a pod is in terminating state for > 5 mins, it's eligible for GC
DEFAULT_STATUS_THRESHOLD=2 # Status threshold in minutes for problematic pods

#  We have checked all the pre-requisites by now, so we'll delete Spark application and its dependencies
delete_spark_app() {
    local app_name=$1
    local app_namespace=$2

    echo "Deleting Spark application $app_name and its dependencies..."

    # Delete the Spark application
    echo "Deleting the Spark app $app_name"
    kubectl delete sparkapplication "$app_name" -n "$app_namespace"
    # TODO: upgrade the script to be only executed when a spark-app has the corresponding dependency on NFS
    delete_nfs_dependencies "$app_name" "$app_namespace"
}

# Delete NFS dependencies
delete_nfs_dependencies() {
  local app_name=$1
  local app_namespace=$2
  # Delete the corresponding PVCs
  PVC_NAME="$app_name-pvc"
  PV_NAME=$(kubectl get pvc "$PVC_NAME" -n "$app_namespace" -o jsonpath='{.spec.volumeName}')
  NFS_SERVER_NAME="$app_name-nfs-server"
  POD_NAME=$(kubectl get pods -n "$app_namespace" -l "app=$NFS_SERVER_NAME" -o jsonpath='{.items[0].metadata.name}')

  if [ -n "$PV_NAME" ]; then
    echo "Deleting pv: $PV_NAME"
    kubectl patch pv $PV_NAME -p '{"metadata":{"finalizers": []}}'
    kubectl delete pv $PV_NAME
    echo "Deleting /exports/$PV_NAME directory from the pod $POD_NAME"
    kubectl exec "$POD_NAME" -n "$app_namespace" -- /bin/sh -c "rm -rf '/exports/$PV_NAME'"
  else
    echo "PV $PV_NAME not present"
  fi

  echo "Deleting the corresponding PVC $PVC_NAME"
  if kubectl get pvc "$PVC_NAME" -n "$app_namespace" > /dev/null 2>&1; then
    kubectl patch pvc "$PVC_NAME" -p '{"metadata":{"finalizers": []}}'
    kubectl delete pvc "$PVC_NAME"
  else
    echo "PVC $PVC_NAME does not exist"
  fi

  # PV deletion is take care before, delete the pv directory from node now
  if [ -n "$PV_NAME" ]; then
    printf "\nDeleting $PV_NAME directory from the pod $POD_NAME, which will delete it from the node"
    kubectl exec "$POD_NAME" -- /bin/sh -c "rm -rf '/exports/$PV_NAME'"
  else
    printf "PV $PV_NAME does not exist"
  fi

  # Delete the NFS deployment, service, and storage class
  echo "Deleting the NFS storage class $NFS_SERVER_NAME"
  kubectl delete storageclass "$NFS_SERVER_NAME-$app_namespace"
  echo "Deleting the NFS deployment $NFS_SERVER_NAME"
  kubectl delete deployment "$NFS_SERVER_NAME" -n "$app_namespace"
  echo "Deleting the NFS service $NFS_SERVER_NAME"
  kubectl delete service "$NFS_SERVER_NAME" -n "$app_namespace"
}

# takes 2 arguments, pod_name and pod_creation_timestamp
# deletes the pod if it's been running for more than the threshold, default being 2 mins
delete_pod() {
  POD_NAME=$1
  POD_CREATION_TIME=$(date -d "$2" +%s)
  CURRENT_TIME=$(date +%s)
  POD_AGE=$(((CURRENT_TIME - POD_CREATION_TIME) / 60))
  local POD_TTL=${3:-$POD_TERMINATING_STATE_THRESHOLD}
  printf "\nReceived pod deletion request"
  # Check if the pod age is greater than the age threshold
  if [ "$POD_AGE" -gt "$POD_TTL" ]; then
    echo "Deleting pod: $POD_NAME (Age: $POD_AGE seconds)"
    kubectl delete pod "$POD_NAME" --grace-period=0
  fi
}

# Restarting the driver pods as the long running ones can cause variety of issues, one of them being
# tokenExpired exceptions, we don't want to delete any dependencies of this app, but only restart this driver pod
recycle_driver_pods() {
  printf "\nRecycling driver pods as these can get stale and cause tokenExpirations when connecting to cloud storage"
  kubectl get pods -o custom-columns=NAME:.metadata.name,AGE:.metadata.creationTimestamp --no-headers | grep driver | while read -r POD_NAME POD_CREATION_TIMESTAMP; do
    delete_pod "$POD_NAME" "$POD_CREATION_TIMESTAMP" "$POD_AGE_THRESHOLD"
  done
}

# terminate driver pods that don't have any backing spark-app
delete_orphan_spark_applications() {
  NAMESPACE=$(kubectl get pods -o custom-columns=NAMESPACE:.metadata.namespace --no-headers | head -n 1 | awk '{print $1}')
  printf "\nTerminating driver pods that don't have the corresponding sparkapps"
  kubectl get pods -n "$NAMESPACE" -o custom-columns=NAME:.metadata.name --no-headers | grep driver | while read -r POD_NAME;  do
    spark_app="${POD_NAME%-driver}"

    if kubectl get sparkapplications "$spark_app" > /dev/null 2>&1; then
      echo "SparkApplication exists for driver pod: $POD_NAME"
    else
      echo "Deleting driver pod without SparkApplication: $POD_NAME"
      kubectl delete pod "$POD_NAME"
      delete_nfs_dependencies "$spark_app" "$NAMESPACE"
    fi
  done
}

# if there's a nfs deployment but no corresponding sparkapp, delete the deployment
delete_orphan_nfs() {
  NAMESPACE=$(kubectl get pods -o custom-columns=NAMESPACE:.metadata.namespace --no-headers | head -n 1 | awk '{print $1}')
  printf "\n Terminating nfs deployments that don't have the corresponding sparkapps"
  kubectl get pods -n "$NAMESPACE" -o custom-columns=NAME:.metadata.name --no-headers | grep driver | while read -r POD_NAME;  do
    # strip out driver from the driver pod name and use it to identify corresponding sparkApp
    spark_app="${POD_NAME%-driver}"

    if kubectl get sparkapplications "$spark_app" > /dev/null 2>&1; then
      echo "SparkApplication exists for driver pod: $POD_NAME"
    else
      echo "Deleting driver pod without SparkApplication: $POD_NAME"
      kubectl delete pod "$POD_NAME" --grace-period=0
      delete_nfs_dependencies "$spark_app" "$NAMESPACE"
    fi
  done
}

# pods can remain in terminating state, delete if it's in such state for a long time
delete_terminating_pods() {
  kubectl get pods -o custom-columns=NAME:.metadata.name,AGE:.metadata.creationTimestamp --no-headers | grep Terminating | while read -r POD_NAME POD_CREATION_TIMESTAMP; do
    delete_pod "$POD_NAME" "$POD_CREATION_TIMESTAMP"
  done
}

#TODO: rollbar integration using CURL?

check_and_cleanup() {
    if [ "$#" -ne 4 ]; then
        echo "Usage: $0 <app_name> <app_namespace> <age_threshold> <status_threshold>"
        exit 1
    fi

    local app_name=$1
    local app_namespace=$2
    local age_threshold=$3
    local status_threshold=$4

    # Validate parameters, we need all
    if [[ -z "$app_name" || -z "$app_namespace" || -z "$age_threshold" || -z "$status_threshold" ]]; then
        echo "Error: All parameters are required."
        echo "Usage: $0 <app_name> <app_namespace> <age_threshold> <status_threshold>"
        exit 1
    fi

    # Get the creation timestamp of the SparkApplication
    printf "\nGetting creation timestamp of SparkApplication:%s" "$app_name"
    APP_CREATION_TIME=$(kubectl get sparkapplication $app_name -n $app_namespace -o jsonpath="{.metadata.creationTimestamp}")

    if [ -z "$APP_CREATION_TIME" ]; then
        echo "SparkApplication $app_name not found."
        exit 1
    fi

    printf "\nThe creation timestamp of the SparkApplication: %s" "$APP_CREATION_TIME"
    APP_CREATION_TIME_UNIX=$(date --date="$APP_CREATION_TIME" +%s)
    printf "\nThe creation time in Unix timestamp: %s" "$APP_CREATION_TIME_UNIX"

    # Get the current Unix timestamp
    NOW_UNIX=$(date -u +%s)
    printf "\n$NOW_UNIX is the current time in Unix timestamp"

    # Calculate the age of the SparkApplication in minutes
    AGE_MINUTES=$(( (NOW_UNIX - APP_CREATION_TIME_UNIX) / 60 ))
    printf "\n$AGE_MINUTES is the age of the SparkApplication in minutes"

    # Determine whether to delete based on SparkApplication age
    if [ $AGE_MINUTES -gt "$age_threshold" ]; then
        printf "age_in minutes: %s" "$AGE_MINUTES"
        delete_spark_app "$app_name" "$app_namespace"
    else
        echo "SparkApplication $app_name is not older than the threshold, but it could have gracefully shutdown"

        # Optionally, check if the driver pod is in a deletable state
        if kubectl get pods "$app_name"-driver -n "$app_namespace" | grep -i -e 'Error' -e 'NotReady' -e 'Pending' -e 'Completed'; then
            printf "\nPod %s has graceful status, cleaning up the rest of the dependencies now!!" "$app_name-driver"

            # Get the start time of the driver pod
            printf "\nGetting start time of:%s" "$app_name-driver"
            DRIVER_START_TIME=$(kubectl get pods "${app_name}"-driver -n "$app_namespace" -o jsonpath="{.metadata.creationTimestamp}")
            printf "\nThe start time of the driver pod: %s" "$DRIVER_START_TIME"
            DRIVER_START_TIME_UNIX=$(date --date="$DRIVER_START_TIME" +%s)
            printf "\nThe start time of the driver pod in Unix timestamp: %s" "$DRIVER_START_TIME_UNIX"

            # Calculate the difference in minutes
            DRIVER_DIFF_MINUTES=$(( (NOW_UNIX - DRIVER_START_TIME_UNIX) / 60 ))
            printf "\nThe difference in minutes for the driver pod: %s" "$DRIVER_DIFF_MINUTES"

            # If the difference is more than STATUS_THRESHOLD, perform deletion
            if [ $DRIVER_DIFF_MINUTES -gt "$status_threshold" ]; then
                delete_spark_app "$app_name" "$app_namespace"
            fi
        fi
    fi
}

periodic_cleanup() {
  # Get the current revision number from the Helm release
  CURRENT_REVISION=$1
  printf "\nCurrent Revision: %s" "$CURRENT_REVISION"

  # List all running SparkApplications in the namespace
  SPARK_APPS=$(kubectl get sparkapplication -o json)
  # The deployer asked the existing app to terminate itself gracefully,
  # but that results in the executors shutting down, but leaving behind driver pod in the Error state,
  # spark application in idle state and PVC unused
  # we'll now clean them up
  REVISIONS=$(echo "$SPARK_APPS" | yq -r '.items[].metadata.labels.revision')
  printf "\nRevisions: %s" "$REVISIONS"
  APP_COUNT=$(echo "$SPARK_APPS" | yq '.items | length')
  printf "\nNumber of spark apps in the namespace: %s" "$APP_COUNT"
  # Get the names of PVs in 'Released' state
  pv_names=$(kubectl get pv -o jsonpath='{.items[?(@.status.phase=="Released")].metadata.name}' | tr ' ' '\n')
  # Delete pvs in released state
  if [[ -n "$pv_names" ]]; then
    while IFS= read -r pv; do
      # Remove the finalizer from the PV and delete it
      printf "\nDeleting pv: %s" "$pv"
      kubectl patch pv $pv -p '{"metadata":{"finalizers": []}}'
      kubectl delete pv $pv
    done <<< "$pv_names"
  else
    printf "\nNo PV names in a released state."
  fi

  local age_threshold=${2:-$DEFAULT_AGE_THRESHOLD}
  local status_threshold=${3:-$DEFAULT_STATUS_THRESHOLD}

  # Filter out the SparkApplications from the current revision
  echo $SPARK_APPS | yq -c --arg CR "$CURRENT_REVISION" '.items[] | select(.metadata.labels.revision != $CR)' | while read -r APP; do
    APP_NAME=$(echo $APP | yq -r '.metadata.name')
    APP_NAMESPACE=$(echo $APP | yq -r '.metadata.namespace')
    printf "\nDealing with the spark-app: %s, in ns: %s" "$APP_NAME" "$APP_NAMESPACE"
    check_and_cleanup "$APP_NAME" "$APP_NAMESPACE" "$age_threshold" "$status_threshold"
  done

  # Delete all pods in 'Error', 'NotReady', 'Completed' state
  printf "\nDeleting any pods that are in the Error, NotReady or Completed state\n"
  kubectl get pods --no-headers | grep -i  -e 'Error' -e 'NotReady' -e 'Completed' -e 'Terminating' | awk '{print $1}' | xargs --no-run-if-empty kubectl delete pod --grace-period=0
  # Delete all jobs that have already completed
  printf "\nDeleting any jobs that are Completed\n"
  kubectl get jobs -o=jsonpath='{.items[?(@.status.succeeded==1)].metadata.name}' | xargs -r kubectl delete job
  delete_orphan_spark_applications
  delete_terminating_pods
  sleep 10
}

send_kill_messages() {
  # Delete all pods in 'Error' state
  printf "\nDeleting any pods that are in the Error state"
  kubectl get pods --no-headers | grep -i 'Error' | awk '{print $1}' | xargs --no-run-if-empty kubectl delete pod --grace-period=0
  # Get the current revision number from the Helm release
  CURRENT_REVISION=$1
  printf "\nCurrent Revision: %s" "$CURRENT_REVISION"

  # List all running SparkApplications in the namespace
  SPARK_APPS=$(kubectl get sparkapplication -o json)

  # Send a message to RabbitMQ for all the old SparkApplications
  RABBITMQ_API_URL=$2
  RABBITMQ_AUTH=$3
  RABBITMQ_PQ=$4
  printf "\nThe rabbit URL: %s" "$RABBITMQ_API_URL"
  # Filter out the SparkApplications from the current revision
  echo $SPARK_APPS | jq -c --arg CR "$CURRENT_REVISION" '.items[] | select(.metadata.labels.revision != $CR)' | while read -r APP; do
    APP_NAME=$(echo $APP | jq -r '.metadata.name')
    APP_NAMESPACE=$(echo $APP | jq -r '.metadata.namespace')
    MSG=$(jq -n --arg APP_NAME "$APP_NAME" '{job: "kill", cluster: $APP_NAME, nonce: -1}')
    escaped_payload=$(echo "$MSG" | jq -sR .)

    printf "\nSending message: $escaped_payload to RabbitMQ for Spark Application: $APP_NAME in namespace: $APP_NAMESPACE"

    # Use curl to send a message to RabbitMQ
    curl -i -u $RABBITMQ_AUTH -H "content-type:application/json" -X POST "$RABBITMQ_API_URL" -d \
    "{
      \"properties\": {
        \"delivery_mode\": 2,
        \"headers\": {}
      },
      \"routing_key\": \"priority.$RABBITMQ_PQ.#\",
      \"payload\": $escaped_payload,
      \"payload_encoding\": \"string\"
    }"
  done
  printf "\nDone deleting all the driver pods that were not of the current version: %s" "$CURRENT_REVISION"
  sleep 10
}

run_cron() {
  echo "Arguments received: $*"
  commandToRun=$1
  currentRevision=$2

  case "$commandToRun" in
    cleaner)
      ageThreshold=$3
      statusThreshold=$4
      printf "\nRunning cleaner job for revision=%s, provided age_threshold=%s and status_threshold=%s" "$currentRevision" "$ageThreshold" "$statusThreshold"
      periodic_cleanup "$currentRevision" "$ageThreshold" "$statusThreshold"
      # Recycle driver pods periodically
      printf "\nRecycle driver pods if eligible"
      recycle_driver_pods
      ;;
    deployer)
      rabbitMQURL=$3
      rabbitMQAuth=$4
      rabbitMQSubSystem=$5
      printf "\nRunning deployer job for revision=%s, rabbitMQURL=%s rabbitMQAuth=%s and rabbitMQSubSystem=%s" "$currentRevision" "$rabbitMQURL" "****" "$rabbitMQSubSystem"
      send_kill_messages "$currentRevision" "$rabbitMQURL" "$rabbitMQAuth" "$rabbitMQSubSystem"
      ;;
    *)
      # commands to execute if $variable does not match any pattern
      ;;
  esac
  sleep 10
}

# Run the periodic_cleanup function with parameters
run_cron "$@"
